use strict;
use diagnostics;
use warnings;

my @steps = (1, 2, 4);
my @stepProb = (0.5, 0.25, 0.25);
my $maxTeeth = 0;
my $grandSumOfProbs = 0;
my $numMatchingPermutations = 0;
my $desiredNumSteps = 0;
my $printPermutation = 0;
my %historyHash;

if (scalar (@ARGV) != 3)
{
    printf "Incorrect number of command line options\n";
    printf "Usage: Calculate_MM2_Prob.pl <desired number of teeth >\n";
    printf "                             <desired number of steps>\n";
    printf "                             <print permutations?>\n";
    exit (-1);
}

$maxTeeth = $ARGV[0];
$desiredNumSteps = $ARGV[1];
$printPermutation = $ARGV[2];

printf "-------------------------------------------------------------------\n";
printf "\n";
printf "Determine probabilities for %d teeth, %d steps\n", $maxTeeth, $desiredNumSteps;
printf "\n";

findPermutations (0, $desiredNumSteps, 0, 1, ());

printf "\n";
printf "Total # permutations: %20.0f\n", $numMatchingPermutations;
printf "Total sum of probabilities: %20.18f\n", $grandSumOfProbs; 

sub findPermutations
{
    my ($depth, $maxDepth, $sumTeeth, $prob, @history) = @_;
    my $i;
    
    if (($sumTeeth == $maxTeeth) && ($depth == $desiredNumSteps))
    {
        # If we encountered the hash signature, skip this permutation.
        if (!hashEncountered (@history))
        {
            my $numPermutations = numberPermutations (@history);

            if ($printPermutation != 0)
            {
                printf "%2d / %20.18f (permutations: %20.0f): ", $sumTeeth, $prob, $numPermutations;
                for ($i=0; $i<scalar(@history); ++$i)
                {
                    printf "%1d ", $history[$i];
                }
                printf "\n";
            }

            $grandSumOfProbs += $prob * $numPermutations;
            $numMatchingPermutations += $numPermutations;
        }
        
        return;
    }
    
    if ($depth >= $maxDepth)
    {
        return;
    }
    
    if ((($maxTeeth - $sumTeeth) / ($maxDepth - $depth)) > $steps [$#steps])
    {
        return;
    }
    
    for ($i=0; $i<scalar(@steps); ++$i)
    {
        if ($depth > 1)
        {
            # Make sure the new step is not less than the previous one.  We'll use the
            # formula for combinations to calculate the probabilities.
            if ($steps[$i] < $history[$depth-1])
            {
                next;
            }
        }
        
        $history[$depth] = $steps[$i];
        findPermutations ($depth+1, $maxDepth, $sumTeeth+$steps[$i], $prob * $stepProb[$i], @history);
    }
}

sub hashEncountered
{
    # Create hash key of @history
    my (@history) = @_;
    my @temp = sort @history;
    my $key = join "-", @temp;
    my $result = 0;
    
    if (defined ($historyHash {$key}))
    {
        $result = 1;
    }
    
    $historyHash {$key} = 1;
    
    return $result;
}

sub numberPermutations
{
    my (@history) = @_;
    my %freq;
    my $i;
    my $temp;
    my $result = 0;
    my @numerator;
    my @denominator;
    my $j;
    
    for ($i=0; $i<scalar(@history); ++$i)
    {
        if (defined ($freq{$history[$i]}))
        {
            ++$freq{$history[$i]};
        }
        else
        {
            $freq{$history[$i]} = 1;
        }
    }

    push (@numerator, (1..scalar(@history)));
    
    foreach $temp (keys %freq)
    {
        push (@denominator, (1..$freq{$temp}));
    }
    
    # Cancel identical values
    my $continue = 1;
    while ($continue != 0)
    {
        $continue = 0;

        #printf "Numerator:  %s\n", join ",", @numerator;
        #printf "Denominator:%s\n", join ",", @denominator;
        #printf "\n";
        
        for ($i=0; $i<scalar(@numerator); ++$i)
        {
            if ($numerator[$i] == 1)
            {
                next;
            }
            
            for ($j=0; $j<scalar(@denominator); ++$j)
            {
                if ($denominator[$j] == 1)
                {
                    next;
                }
                
                my $g = gcd($denominator[$j], $numerator[$i]);
                
                if  (
                        ($denominator[$j] >= $numerator[$i])
                        &&
                        ($g != 1)
                    )
                {
                    $continue = 1;
                    $denominator[$j] /= $g;
                    $numerator[$i]   /= $g;
                    last;
                }
                
                if  (
                        ($numerator[$i] >= $denominator[$j])
                        &&
                        ($g != 1)
                    )
                {
                    $continue = 1;
                    $numerator[$i]   /= $g;
                    $denominator[$j] /= $g;
                    last;
                }
            }
            
            if ($continue == 1)
            {
                # Begin the <i> loop from the beginning
                last;
            }
        }
    }
    
    $result = 1;
    
    for ($i=0; $i<scalar(@numerator); ++$i)
    {
        $result *= $numerator[$i];
    }
    
    for ($i=0; $i<scalar(@denominator); ++$i)
    {
        $result /= $denominator[$i];
    }
    
    return $result;
}

sub factorial
{
    my ($n) = @_;
    my $i;
    my $result = 1;
    
    for ($i=2; $i<=$n; ++$i)
    {
        $result *= $i;
    }
    
    return $result;
}

sub gcd
{
    my ($a, $b) = @_;
    
    if ($a < $b)
    {
        # Swap
        my $temp = $a;
        
        $a = $b;
        $b = $temp;
    }
    
    # Computation
    while ($b != 0)
    {
        my $t = $b;
        $b = $a % $b;
        $a = $t;
    }
    
    return $a;
}
